@@ -14,10 +14,7 @@ use Scalar::Util qw/blessed weaken/;
 
 # "Robots don't have any emotions, and sometimes that makes me very sad."
 has controller_class => 'Mojolicious::Controller';
-has mode             => sub { ($ENV{MOJO_MODE} || 'development') };
-has on_process       => sub {
-  sub { shift->dispatch(@_) }
-};
+has mode => sub { ($ENV{MOJO_MODE} || 'development') };
 has plugins  => sub { Mojolicious::Plugins->new };
 has renderer => sub { Mojolicious::Renderer->new };
 has routes   => sub { Mojolicious::Routes->new };
@@ -35,7 +32,7 @@ has static   => sub { Mojolicious::Static->new };
 has types    => sub { Mojolicious::Types->new };
 
 our $CODENAME = 'Leaf Fluttering In Wind';
-our $VERSION  = '2.23';
+our $VERSION  = '2.57';
 
 # "These old doomsday devices are dangerously unstable.
 #  I'll rest easier not knowing where they are."
@@ -62,10 +59,10 @@ sub DESTROY { }
 sub new {
   my $self = shift->SUPER::new(@_);
 
-  # Root directories
+  # Paths
   my $home = $self->home;
-  $self->renderer->root($home->rel_dir('templates'));
-  $self->static->root($home->rel_dir('public'));
+  push @{$self->renderer->paths}, $home->rel_dir('templates');
+  push @{$self->static->paths},   $home->rel_dir('public');
 
   # Default to application namespace
   my $r = $self->routes;
@@ -76,7 +73,7 @@ sub new {
   $r->hide(qw/param redirect_to render render_content render_data/);
   $r->hide(qw/render_exception render_json render_not_found render_partial/);
   $r->hide(qw/render_static render_text rendered req res respond_to/);
-  $r->hide(qw/send_message session signed_cookie stash tx ua url_for write/);
+  $r->hide(qw/send session signed_cookie stash tx ua url_for write/);
   $r->hide('write_chunk');
 
   # Prepare log
@@ -113,28 +110,8 @@ sub build_tx {
   return $tx;
 }
 
-# "Amy, technology isn't intrinsically good or evil. It's how it's used.
-#  Like the Death Ray."
-sub defaults {
-  my $self = shift;
+sub defaults { shift->_dict(defaults => @_) }
 
-  # Hash
-  $self->{defaults} ||= {};
-  return $self->{defaults} unless @_;
-
-  # Get
-  return $self->{defaults}->{$_[0]} unless @_ > 1 || ref $_[0];
-
-  # Set
-  my $values = ref $_[0] ? $_[0] : {@_};
-  for my $key (keys %$values) {
-    $self->{defaults}->{$key} = $values->{$key};
-  }
-
-  return $self;
-}
-
-# The default dispatchers with exception handling
 sub dispatch {
   my ($self, $c) = @_;
 
@@ -154,10 +131,7 @@ sub dispatch {
   return if $res->code;
   if (my $code = ($tx->req->error)[1]) { $res->code($code) }
   elsif ($tx->is_websocket) { $res->code(426) }
-  unless ($self->routes->dispatch($c)) {
-    $c->render_not_found
-      unless $res->code;
-  }
+  $c->render_not_found unless $self->routes->dispatch($c) || $res->code;
 }
 
 # "Bite my shiny metal ass!"
@@ -171,14 +145,27 @@ sub handler {
     $tx    = $tx->tx;
   }
 
-  # Build default controller and process
+  # Build default controller
   my $defaults = $self->defaults;
   @{$stash}{keys %$defaults} = values %$defaults;
   my $c =
     $self->controller_class->new(app => $self, stash => $stash, tx => $tx);
   weaken $c->{app};
   weaken $c->{tx};
-  unless (eval { $self->on_process->($self, $c); 1 }) {
+
+  # Dispatcher
+  unless ($self->{dispatch}) {
+    $self->hook(
+      around_dispatch => sub {
+        my ($next, $c) = @_;
+        $c->app->dispatch($c);
+      }
+    );
+    $self->{dispatch}++;
+  }
+
+  # Process
+  unless (eval { $self->plugins->emit_chain(around_dispatch => $c) }) {
     $self->log->fatal("Processing request failed: $@");
     $tx->res->code(500);
     $tx->resume;
@@ -189,12 +176,9 @@ sub handler {
     unless $stash->{'mojo.rendered'} || $tx->is_writing;
 }
 
-# "This snow is beautiful. I'm glad global warming never happened.
-#  Actually, it did. But thank God nuclear winter canceled it out."
 sub helper {
-  my $self = shift;
-  my $name = shift;
-  my $r    = $self->renderer;
+  my ($self, $name) = (shift, shift);
+  my $r = $self->renderer;
   $self->log->debug(qq/Helper "$name" already exists, replacing./)
     if exists $r->helpers->{$name};
   $r->add_helper($name, @_);
@@ -214,7 +198,6 @@ sub plugin {
   $self->plugins->register_plugin(shift, $self, @_);
 }
 
-# Start command system
 sub start {
   my $class = shift;
 
@@ -228,7 +211,6 @@ sub start {
   Mojolicious::Commands->start(@_);
 }
 
-# This will run once at startup
 sub startup { }
 
 1;
@@ -283,10 +265,9 @@ L<Mojolicious::Controller>.
   $app     = $app->mode('production');
 
 The operating mode for your application, defaults to the value of the
-C<MOJO_MODE> environment variable or C<development>.
-You can also add per mode logic to your application by defining methods named
-C<${mode}_mode> in the application class, which will be called right before
-C<startup>.
+C<MOJO_MODE> environment variable or C<development>. You can also add per
+mode logic to your application by defining methods named C<${mode}_mode> in
+the application class, which will be called right before C<startup>.
 
   sub development_mode {
     my $self = shift;
@@ -302,31 +283,14 @@ Right before calling C<startup> and mode specific methods, L<Mojolicious>
 will pick up the current mode, name the log file after it and raise the log
 level from C<debug> to C<info> if it has a value other than C<development>.
 
-=head2 C<on_process>
-
-  my $process = $app->on_process;
-  $app        = $app->on_process(sub {...});
-
-Request processing callback, defaults to calling the C<dispatch> method.
-Generally you will use a plugin or controller instead of this, consider it
-the sledgehammer in your toolbox.
-
-  my $next = $app->on_process;
-  $app->on_process(sub {
-    my ($self, $c) = @_;
-    return $c->render(text => 'Hello world!')
-      if $c->req->url->path->contains('/hello');
-    $self->$next($c);
-  });
-
 =head2 C<plugins>
 
   my $plugins = $app->plugins;
   $app        = $app->plugins(Mojolicious::Plugins->new);
 
-The plugin loader, defaults to a L<Mojolicious::Plugins> object.
-You can usually leave this alone, see L<Mojolicious::Plugin> if you want to
-write a plugin or the C<plugin> method below if you want to load a plugin.
+The plugin loader, defaults to a L<Mojolicious::Plugins> object. You can
+usually leave this alone, see L<Mojolicious::Plugin> if you want to write a
+plugin or the C<plugin> method below if you want to load a plugin.
 
 =head2 C<renderer>
 
@@ -334,18 +298,20 @@ write a plugin or the C<plugin> method below if you want to load a plugin.
   $app         = $app->renderer(Mojolicious::Renderer->new);
 
 Used in your application to render content, defaults to a
-L<Mojolicious::Renderer> object.
-The two main renderer plugins L<Mojolicious::Plugin::EPRenderer> and
-L<Mojolicious::Plugin::EPLRenderer> contain more information.
+L<Mojolicious::Renderer> object. The two main renderer plugins
+L<Mojolicious::Plugin::EPRenderer> and L<Mojolicious::Plugin::EPLRenderer>
+contain more information.
+
+  # Add another "templates" directory
+  push @{$app->renderer->paths}, '/foo/bar/templates';
 
 =head2 C<routes>
 
   my $routes = $app->routes;
   $app       = $app->routes(Mojolicious::Routes->new);
 
-The routes dispatcher, defaults to a L<Mojolicious::Routes> object.
-You use this in your startup method to define the url endpoints for your
-application.
+The routes dispatcher, defaults to a L<Mojolicious::Routes> object. You use
+this in your startup method to define the url endpoints for your application.
 
   sub startup {
     my $self = shift;
@@ -360,9 +326,9 @@ application.
   $app       = $app->secret('passw0rd');
 
 A secret passphrase used for signed cookies and the like, defaults to the
-application name which is not very secure, so you should change it!!!
-As long as you are using the unsecure default there will be debug messages in
-the log file reminding you to change your passphrase.
+application name which is not very secure, so you should change it!!! As long
+as you are using the unsecure default there will be debug messages in the log
+file reminding you to change your passphrase.
 
 =head2 C<sessions>
 
@@ -370,18 +336,21 @@ the log file reminding you to change your passphrase.
   $app         = $app->sessions(Mojolicious::Sessions->new);
 
 Simple signed cookie based sessions, defaults to a L<Mojolicious::Sessions>
-object.
-You can usually leave this alone, see L<Mojolicious::Controller/"session">
-for more information about working with session data.
+object. You can usually leave this alone, see
+L<Mojolicious::Controller/"session"> for more information about working with
+session data.
 
 =head2 C<static>
 
   my $static = $app->static;
   $app       = $app->static(Mojolicious::Static->new);
 
-For serving static assets from your C<public> directory, defaults to a
+For serving static files from your C<public> directories, defaults to a
 L<Mojolicious::Static> object.
 
+  # Add another "public" directory
+  push @{$app->static->paths}, '/foo/bar/public';
+
 =head2 C<types>
 
   my $types = $app->types;
@@ -401,10 +370,9 @@ new ones.
 
   my $app = Mojolicious->new;
 
-Construct a new L<Mojolicious> application.
-Will automatically detect your home directory and set up logging based on
-your current operating mode.
-Also sets up the renderer, static dispatcher and a default set of plugins.
+Construct a new L<Mojolicious> application. Will automatically detect your
+home directory and set up logging based on your current operating mode. Also
+sets up the renderer, static dispatcher and a default set of plugins.
 
 =head2 C<build_tx>
 
@@ -420,7 +388,8 @@ object.
   $app         = $app->defaults({foo => 'bar'});
   $app         = $app->defaults(foo => 'bar');
 
-Default values for the stash, assigned for every new request.
+Default values for L<Mojolicious::Controller/"stash">, assigned for every new
+request.
 
   $app->defaults->{foo} = 'bar';
   my $foo = $app->defaults->{foo};
@@ -430,13 +399,14 @@ Default values for the stash, assigned for every new request.
 
   $app->dispatch($c);
 
-The heart of every Mojolicious application, calls the static and routes
+The heart of every Mojolicious application, calls the C<static> and C<routes>
 dispatchers for every request and passes them a L<Mojolicious::Controller>
 object.
 
 =head2 C<handler>
 
-  $tx = $app->handler($tx);
+  $app->handler($tx);
+  $app->handler($c);
 
 Sets up the default controller and calls process for every request.
 
@@ -460,59 +430,88 @@ and the application object, as well as a function in C<ep> templates.
 
   $app->hook(after_dispatch => sub {...});
 
-Extend L<Mojolicious> by adding hooks.
+Extend L<Mojolicious> with hooks.
+
+  # Dispatchers will not run if there's already a response code defined
+  $app->hook(before_dispatch => sub {
+    my $c = shift;
+    $c->render(text => 'Skipped dispatchers!')
+      if $c->req->url->path->contains('/do_not_dispatch');
+  });
 
 These hooks are currently available and are emitted in the listed order:
 
 =over 2
 
-=item after_build_tx
+=item B<after_build_tx>
 
-Emitted in reverse order right after the transaction is built and before the
-HTTP request gets parsed.
+Emitted right after the transaction is built and before the HTTP request gets
+parsed.
 
   $app->hook(after_build_tx => sub {
     my ($tx, $app) = @_;
+    ...
   });
 
-One use case would be upload progress bars.
-(Passed the transaction and application instances)
+This is a very powerful hook and should not be used lightly, it makes some
+rather advanced features such as upload progress bars possible, just note
+that it will not work for embedded applications. (Passed the transaction and
+application instances)
 
-=item before_dispatch
+=item B<before_dispatch>
 
 Emitted right before the static and routes dispatchers start their work.
 
   $app->hook(before_dispatch => sub {
-    my $self = shift;
+    my $c = shift;
+    ...
   });
 
 Very useful for rewriting incoming requests and other preprocessing tasks.
 (Passed the default controller instance)
 
-=item after_static_dispatch
+=item B<after_static_dispatch>
 
 Emitted in reverse order after the static dispatcher determined if a static
 file should be served and before the routes dispatcher starts its work.
 
   $app->hook(after_static_dispatch => sub {
-    my $self = shift;
+    my $c = shift;
+    ...
   });
 
 Mostly used for custom dispatchers and postprocessing static file responses.
 (Passed the default controller instance)
 
-=item after_dispatch
+=item B<after_dispatch>
 
-Emitted in reverse order after a response has been rendered.
-Note that this hook can trigger before C<after_static_dispatch> due to its
-dynamic nature.
+Emitted in reverse order after a response has been rendered. Note that this
+hook can trigger before C<after_static_dispatch> due to its dynamic nature.
 
   $app->hook(after_dispatch => sub {
-    my $self = shift;
+    my $c = shift;
+    ...
   });
 
-Useful for all kinds of postprocessing tasks.
-(Passed the current controller instance)
+Useful for all kinds of postprocessing tasks. (Passed the current controller
+instance)
+
+=item B<around_dispatch>
+
+Emitted right before the C<before_dispatch> hook and wraps around the whole
+dispatch process, so you have to manually forward to the next hook if you
+want to continue the chain.
+
+  $app->hook(around_dispatch => sub {
+    my ($next, $c) = @_;
+    ...
+    $next->();
+    ...
+  });
+
+This is a very powerful hook and should not be used lightly, consider it the
+sledgehammer in your toolbox. (Passed a closure leading to the next hook and
+the current controller instance)
 
 =back
 
@@ -605,17 +604,17 @@ startup.
 
   sub startup {
     my $self = shift;
+    ...
   }
 
 =head1 HELPERS
 
 In addition to the attributes and methods above you can also call helpers on
-instances of L<Mojolicious>.
-This includes all helpers from L<Mojolicious::Plugin::DefaultHelpers> and
-L<Mojolicious::Plugin::TagHelpers>.
-Note that application helpers are always called with a new
-C<controller_class> instance, so they can't depend on or change controller
-state, which includes request, response and stash.
+instances of L<Mojolicious>. This includes all helpers from
+L<Mojolicious::Plugin::DefaultHelpers> and
+L<Mojolicious::Plugin::TagHelpers>. Note that application helpers are always
+called with a new C<controller_class> instance, so they can't depend on or
+change controller state, which includes request, response and stash.
 
   $app->log->debug($app->dumper({foo => 'bar'}));
 
@@ -623,21 +622,21 @@ state, which includes request, response and stash.
 
 =head2 Web
 
-  http://mojolicio.us
+L<http://mojolicio.us>
 
 =head2 IRC
 
-  #mojo on irc.perl.org
+C<#mojo> on C<irc.perl.org>
 
 =head2 Mailing-List
 
-  http://groups.google.com/group/mojolicious
+L<http://groups.google.com/group/mojolicious>
 
 =head1 DEVELOPMENT
 
 =head2 Repository
 
-  http://github.com/kraih/mojo
+L<http://github.com/kraih/mojo>
 
 =head1 BUNDLED FILES
 
@@ -646,7 +645,7 @@ directory.
 
 =head2 Mojolicious Artwork
 
-  Copyright (C) 2010-2011, Sebastian Riedel.
+  Copyright (C) 2010-2012, Sebastian Riedel.
 
 Licensed under the CC-SA License, Version 3.0
 L<http://creativecommons.org/licenses/by-sa/3.0>.
@@ -685,9 +684,23 @@ that have been used in the past.
 
 0.999920, C<Snowman> (u2603)
 
-=head1 AUTHOR
+=head1 PROJECT FOUNDER
 
-Sebastian Riedel, C<sri@cpan.org>.
+Sebastian Riedel, C<sri@cpan.org>
+
+=head1 CORE DEVELOPERS
+
+Current members of the core team in alphabetical order:
+
+=over 4
+
+Abhijit Menon-Sen, C<ams@cpan.org>
+
+Glen Hinkle, C<tempire@cpan.org>
+
+Marcus Ramberg, C<mramberg@cpan.org>
+
+=back
 
 =head1 CREDITS
 
@@ -695,8 +708,6 @@ In alphabetical order:
 
 =over 2
 
-Abhijit Menon-Sen
-
 Adam Kennedy
 
 Adriano Ferreira
@@ -711,6 +722,8 @@ Anatoly Sharifulin
 
 Andre Vieth
 
+Andreas Jaekel
+
 Andrew Fresh
 
 Andreas Koenig
@@ -727,6 +740,10 @@ Audrey Tang
 
 Ben van Staveren
 
+Benjamin Erhart
+
+Bernhard Graf
+
 Breno G. de Oliveira
 
 Brian Duggan
@@ -751,16 +768,18 @@ Danijel Tasov
 
 David Davis
 
+Diego Kuperman
+
 Dmitriy Shalashov
 
 Dmitry Konstantinov
 
+Douglas Christopher Wilson
+
 Eugene Toropov
 
 Gisle Aas
 
-Glen Hinkle
-
 Graham Barr
 
 Henry Tang
@@ -775,6 +794,8 @@ Jaroslav Muhin
 
 Jesse Vincent
 
+Johannes Plunien
+
 John Kingsley
 
 Jonathan Yu
@@ -793,8 +814,6 @@ Magnus Holm
 
 Maik Fischer
 
-Marcus Ramberg
-
 Mark Stosberg
 
 Matthew Lineen
@@ -811,6 +830,8 @@ Mons Anderson
 
 Moritz Lenz
 
+Neil Watkiss
+
 Nils Diewald
 
 Oleg Zhelo
@@ -853,6 +874,8 @@ Skye Shaw
 
 Stanis Trendelenburg
 
+Stephane Este-Gracias
+
 Tatsuhiko Miyagawa
 
 Terrence Brannon
@@ -873,6 +896,8 @@ Victor Engmark
 
 Viliam Pucik
 
+Wes Cravens
+
 Yaroslav Korshak
 
 Yuki Kimoto
@@ -883,7 +908,7 @@ Zak B. Elep
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2008-2011, Sebastian Riedel.
+Copyright (C) 2008-2012, Sebastian Riedel.
 
 This program is free software, you can redistribute it and/or modify it under
 the terms of the Artistic License version 2.0.
