@@ -22,19 +22,18 @@ has match => sub {
 has tx => sub { Mojo::Transaction::HTTP->new };
 
 # Bundled files
-my $H = Mojo::Home->new;
+our $H = Mojo::Home->new;
 $H->parse($H->parse($H->mojo_lib_dir)->rel_dir('Mojolicious/templates'));
-our $EXCEPTION = $H->slurp_rel_file('exception.html.ep');
-our $DEVELOPMENT_EXCEPTION =
-  $H->slurp_rel_file('exception.development.html.ep');
-our $NOT_FOUND = $H->slurp_rel_file('not_found.html.ep');
-our $DEVELOPMENT_NOT_FOUND =
-  $H->slurp_rel_file('not_found.development.html.ep');
+our $EXCEPTION     = $H->slurp_rel_file('exception.html.ep');
+our $DEV_EXCEPTION = $H->slurp_rel_file('exception.development.html.ep');
+our $MOJOBAR       = $H->slurp_rel_file('mojobar.html.ep');
+our $NOT_FOUND     = $H->slurp_rel_file('not_found.html.ep');
+our $DEV_NOT_FOUND = $H->slurp_rel_file('not_found.development.html.ep');
 
 # Reserved stash values
 my @RESERVED = (
-  qw/action app cb class controller data exception extends format handler/,
-  qw/json layout method namespace partial path status template text/
+  qw/action app cb controller data extends format handler json layout/,
+  qw/namespace partial path status template text/
 );
 my %RESERVED;
 $RESERVED{$_}++ for @RESERVED;
@@ -130,10 +129,6 @@ sub flash {
   $flash = {} unless $flash && ref $flash eq 'HASH';
   $session->{new_flash} = $flash;
 
-  # DEPRECATED in Smiling Face With Sunglasses!
-  warn "Direct hash access to the flash is DEPRECATED!!!\n" and return $flash
-    unless @_;
-
   # Set
   my $values = @_ > 1 ? {@_} : $_[0];
   $session->{new_flash} = {%$flash, %$values};
@@ -149,41 +144,39 @@ sub on {
   $tx->on($name => sub { shift and $self->$cb(@_) });
 }
 
-# DEPRECATED in Leaf Fluttering In Wind!
-sub on_finish {
-  warn <<EOF;
-Mojolicious::Controller->on_finish is DEPRECATED in favor of
-Mojolicious::Controller->on!
-EOF
-  shift->on(finish => @_);
-}
-
 # "Just make a simple cake. And this time, if someone's going to jump out of
 #  it make sure to put them in *after* you cook it."
 sub param {
-  my $self = shift;
-  my $name = shift;
+  my ($self, $name) = (shift, shift);
 
-  # List
-  my $p = $self->stash->{'mojo.captures'} || {};
+  # List names
+  my $p = $self->stash->{'mojo.captures'} ||= {};
+  my $req = $self->req;
   unless (defined $name) {
     my %seen;
-    my @keys = grep { !$seen{$_}++ } $self->req->param;
+    my @keys = grep { !$seen{$_}++ } $req->param;
+    push @keys, grep { !$seen{$_}++ } map { $_->name } @{$req->uploads};
     push @keys, grep { !$RESERVED{$_} && !$seen{$_}++ } keys %$p;
     return sort @keys;
   }
 
-  # Override value
+  # Override values
   if (@_) {
-    $p->{$name} = $_[0];
+    $p->{$name} = @_ > 1 ? [@_] : $_[0];
     return $self;
   }
 
-  # Captured unreserved value
-  return $p->{$name} if !$RESERVED{$name} && exists $p->{$name};
+  # Captured unreserved values
+  if (!$RESERVED{$name} && defined(my $v = $p->{$name})) {
+    return ref $v && ref $v eq 'ARRAY' ? wantarray ? @$v : $$v[0] : $v;
+  }
 
-  # Param value
-  return $self->req->param($name);
+  # Upload
+  my $upload = $req->upload($name);
+  return $upload if $upload;
+
+  # Param values
+  return $req->param($name);
 }
 
 # "Is there an app for kissing my shiny metal ass?
@@ -234,7 +227,7 @@ sub render {
   # Render
   my ($output, $type) = $self->app->renderer->render($self, $args);
   return unless defined $output;
-  return $output if $args->{partial};
+  return Mojo::ByteStream->new($output) if $args->{partial};
 
   # Prepare response
   my $res = $self->res;
@@ -305,26 +298,16 @@ sub render_exception {
     exception        => $e,
     'mojo.exception' => 1
   };
-  my $inline = $mode eq 'development' ? $DEVELOPMENT_EXCEPTION : $EXCEPTION;
+  my $inline = $mode eq 'development' ? $DEV_EXCEPTION : $EXCEPTION;
   return if $self->_render_fallbacks($options, 'exception', $inline);
   $options->{format} = 'html';
   $self->_render_fallbacks($options, 'exception', $inline);
 }
 
-# DEPRECATED in Smiling Face With Sunglasses!
-sub render_inner {
-  warn <<EOF;
-Mojolicious::Controller->render_inner is DEPRECATED in favor of
-Mojolicious::Controller->render_content!
-EOF
-  shift->render_content(@_);
-}
-
 # "If you hate intolerance and being punched in the face by me,
 #  please support Proposition Infinity."
 sub render_json {
-  my $self = shift;
-  my $json = shift;
+  my ($self, $json) = (shift, shift);
   my $args = ref $_[0] ? $_[0] : {@_};
   $args->{json} = $json;
   return $self->render($args);
@@ -343,22 +326,15 @@ sub render_not_found {
   return if $stash->{'mojo.exception'};
   return if $stash->{'mojo.not_found'};
 
-  # Check for POD plugin
-  my $guide =
-      $self->app->renderer->helpers->{pod_to_html}
-    ? $self->url_for('/perldoc')
-    : 'http://mojolicio.us/perldoc';
-
   # Render with fallbacks
   my $mode    = $self->app->mode;
   my $options = {
     template         => "not_found.$mode",
     format           => $stash->{format} || 'html',
     status           => 404,
-    guide            => $guide,
     'mojo.not_found' => 1
   };
-  my $inline = $mode eq 'development' ? $DEVELOPMENT_NOT_FOUND : $NOT_FOUND;
+  my $inline = $mode eq 'development' ? $DEV_NOT_FOUND : $NOT_FOUND;
   return if $self->_render_fallbacks($options, 'not_found', $inline);
   $options->{format} = 'html';
   $self->_render_fallbacks($options, 'not_found', $inline);
@@ -369,23 +345,19 @@ sub render_not_found {
 sub render_partial {
   my $self     = shift;
   my $template = @_ % 2 ? shift : undef;
-  my $args     = {@_};
-
+  my $args     = {@_, partial => 1};
   $args->{template} = $template if defined $template;
-  $args->{partial} = 1;
-
-  return Mojo::ByteStream->new($self->render($args));
+  return $self->render($args);
 }
 
 sub render_static {
   my ($self, $file) = @_;
 
   my $app = $self->app;
-  unless ($app->static->serve($self, $file)) {
-    $app->log->debug(
-      qq/Static file "$file" not found, public directory missing?/);
-    return;
-  }
+  $app->log->debug(
+    qq/Static file "$file" not found, public directory missing?/)
+    and return
+    unless $app->static->serve($self, $file);
   $self->rendered;
 
   return 1;
@@ -455,13 +427,13 @@ sub respond_to {
   ref $target eq 'CODE' ? $target->($self) : $self->render($target);
 }
 
-sub send_message {
+sub send {
   my ($self, $message, $cb) = @_;
 
   my $tx = $self->tx;
   Carp::croak('No WebSocket connection to send message to')
     unless $tx->is_websocket;
-  $tx->send_message($message, sub { shift and $self->$cb(@_) if $cb });
+  $tx->send($message, sub { shift and $self->$cb(@_) if $cb });
   $self->rendered(101);
 
   return $self;
@@ -471,22 +443,17 @@ sub send_message {
 sub session {
   my $self = shift;
 
-  # Get
-  my $stash   = $self->stash;
-  my $session = $stash->{'mojo.session'};
-  if ($_[0] && !defined $_[1] && !ref $_[0]) {
-    return unless $session && ref $session eq 'HASH';
-    return $session->{$_[0]};
-  }
-
   # Hash
-  $session = {} unless $session && ref $session eq 'HASH';
-  $stash->{'mojo.session'} = $session;
-  return $session unless @_;
+  my $stash = $self->stash;
+  $stash->{'mojo.session'} ||= {};
+  return $stash->{'mojo.session'} unless @_;
+
+  # Get
+  return $stash->{'mojo.session'}->{$_[0]} unless @_ > 1 || ref $_[0];
 
   # Set
-  my $values = @_ > 1 ? {@_} : $_[0];
-  $stash->{'mojo.session'} = {%$session, %$values};
+  my $values = ref $_[0] ? $_[0] : {@_};
+  $stash->{'mojo.session'} = {%{$stash->{'mojo.session'}}, %$values};
 
   return $self;
 }
@@ -602,7 +569,7 @@ sub url_for {
   # Make path absolute
   my $base_path = $base->path;
   unshift @{$path->parts}, @{$base_path->parts};
-  $base_path->parts([]);
+  $base_path->parts([])->trailing_slash(0);
 
   return $url;
 }
@@ -670,9 +637,8 @@ Mojolicious::Controller - Controller base class
 =head1 DESCRIPTION
 
 L<Mojolicious::Controller> is the base class for your L<Mojolicious>
-controllers.
-It is also the default controller class for L<Mojolicious> unless you set
-C<controller_class> in your application.
+controllers. It is also the default controller class for L<Mojolicious>
+unless you set C<controller_class> in your application.
 
 =head1 ATTRIBUTES
 
@@ -687,6 +653,8 @@ implements the following new ones.
 A reference back to the L<Mojolicious> application that dispatched to this
 controller, defaults to a L<Mojolicious> object.
 
+  $c->app->log->debug('Hello Mojo!');
+
 =head2 C<match>
 
   my $m = $c->match;
@@ -695,6 +663,8 @@ controller, defaults to a L<Mojolicious> object.
 Routes dispatcher results for the current request, defaults to a
 L<Mojolicious::Routes::Match> object.
 
+  my $name = $c->match->endpoint->name;
+
 =head2 C<tx>
 
   my $tx = $c->tx;
@@ -703,6 +673,8 @@ L<Mojolicious::Routes::Match> object.
 The transaction that is currently being processed, usually a
 L<Mojo::Transaction::HTTP> or L<Mojo::Transaction::WebSocket> object.
 
+  my $address = $c->tx->remote_address;
+
 =head1 METHODS
 
 L<Mojolicious::Controller> inherits all methods from L<Mojo::Base> and
@@ -717,6 +689,8 @@ implements the following new ones.
 
 Access request cookie values and create new response cookies.
 
+  my $foo = $c->cookie('foo')->value;
+
 =head2 C<finish>
 
   $c->finish;
@@ -730,14 +704,17 @@ Gracefully end WebSocket connection or long poll stream.
   $c        = $c->flash({foo => 'bar'});
   $c        = $c->flash(foo => 'bar');
 
-Data storage persistent only for the next request, stored in the session.
+Data storage persistent only for the next request, stored in the C<session>.
+
+  $c->flash(message => 'User created successfully!');
+  $c->redirect_to('show_user', id => 23);
 
 =head2 C<on>
 
   my $cb = $c->on(finish => sub {...});
 
-Register event with C<tx>, which is usually a L<Mojo::Transaction::HTTP> or
-L<Mojo::Transaction::WebSocket> object.
+Subscribe to events of C<tx>, which is usually a L<Mojo::Transaction::HTTP>
+or L<Mojo::Transaction::WebSocket> object.
 
   # Emitted when the transaction has been finished
   $c->on(finish => sub {
@@ -757,9 +734,10 @@ L<Mojo::Transaction::WebSocket> object.
   my $foo   = $c->param('foo');
   my @foo   = $c->param('foo');
   $c        = $c->param(foo => 'ba;r');
+  $c        = $c->param(foo => qw/ba;r ba;z/);
 
-Access GET/POST parameters and route captures that are not reserved stash
-values.
+Access GET/POST parameters, file uploads and route captures that are not
+reserved stash values.
 
   # Only GET parameters
   my $foo = $c->req->url->query->param('foo');
@@ -767,6 +745,9 @@ values.
   # Only GET and POST parameters
   my $foo = $c->req->param('foo');
 
+  # Only file uploads
+  my $foo = $c->req->upload('foo');
+
 =head2 C<redirect_to>
 
   $c = $c->redirect_to('named');
@@ -777,8 +758,13 @@ values.
 Prepare a C<302> redirect response, takes the exact same arguments as
 C<url_for>.
 
+  # Conditional redirect
   return $c->redirect_to('login') unless $c->session('user');
 
+  # Moved permanently
+  $c->res->code(301);
+  $c->redirect_to('some_route');
+
 =head2 C<render>
 
   $c->render;
@@ -793,11 +779,9 @@ C<url_for>.
   $c->render('foo/bar', format => 'html');
 
 This is a wrapper around L<Mojolicious::Renderer> exposing pretty much all
-functionality provided by it.
-It will set a default template to use based on the controller and action name
-or fall back to the route name.
-You can call it with a hash of options which can be preceded by an optional
-template name.
+functionality provided by it. It will set a default template to use based on
+the controller and action name or fall back to the route name. You can call
+it with a hash of options which can be preceded by an optional template name.
 
 =head2 C<render_content>
 
@@ -811,8 +795,8 @@ C<extends> features.
 
 =head2 C<render_data>
 
-  $c->render_data($bits);
-  $c->render_data($bits, format => 'png');
+  $c->render_data($bytes);
+  $c->render_data($bytes, format => 'png');
 
 Render the given content as raw bytes, similar to C<render_text> but data
 will not be encoded.
@@ -828,7 +812,7 @@ C<exception.$format.*> and set the response status code to C<500>.
 =head2 C<render_json>
 
   $c->render_json({foo => 'bar'});
-  $c->render_json([1, 2, -3]);
+  $c->render_json([1, 2, -3], status => 201);
 
 Render a data structure as JSON.
 
@@ -836,7 +820,8 @@ Render a data structure as JSON.
 
   $c->render_later;
 
-Disable auto rendering, especially for long polling this can be quite useful.
+Disable automatic rendering, especially for long polling this can be quite
+useful.
 
   $c->render_later;
   Mojo::IOLoop->timer(2 => sub {
@@ -846,7 +831,7 @@ Disable auto rendering, especially for long polling this can be quite useful.
 =head2 C<render_not_found>
 
   $c->render_not_found;
-  $c->render_not_found($resource);
+  $c->render_not_found('some_resource');
 
 Render the not found template C<not_found.$mode.$format.*> or
 C<not_found.$format.*> and set the response status code to C<404>.
@@ -863,8 +848,8 @@ Same as C<render> but returns the rendered result.
   my $success = $c->render_static('images/logo.png');
   my $success = $c->render_static('../lib/MyApp.pm');
 
-Render a static file using L<Mojolicious::Static> relative to the
-C<public> directory of your application.
+Render a static file using L<Mojolicious::Static>, relative to the
+C<public> directories of your application.
 
 =head2 C<render_text>
 
@@ -872,8 +857,8 @@ C<public> directory of your application.
   $c->render_text('Hello World', layout => 'green');
 
 Render the given content as Perl characters, which will be encoded to bytes.
-See C<render_data> for an alternative without encoding.
-Note that this does not change the content type of the response, which is
+See C<render_data> for an alternative without encoding. Note that this does
+not change the content type of the response, which is
 C<text/html;charset=UTF-8> by default.
 
   $c->render_text('Hello World!', format => 'txt');
@@ -885,19 +870,28 @@ C<text/html;charset=UTF-8> by default.
 
 Finalize response and run C<after_dispatch> plugin hook.
 
+  # Stream content directly from file
+  $c->res->content->asset(Mojo::Asset::File->new(path => '/etc/passwd'));
+  $c->res->headers->content_type('text/plain');
+  $c->rendered(200);
+
 =head2 C<req>
 
   my $req = $c->req;
 
-Alias for C<$c-E<gt>tx-E<gt>req>.
-Usually refers to a L<Mojo::Message::Request> object.
+Alias for C<$c-E<gt>tx-E<gt>req>. Usually refers to a
+L<Mojo::Message::Request> object.
+
+  $c->render_json({url => $c->req->url->to_abs->to_string});
 
 =head2 C<res>
 
   my $res = $c->res;
 
-Alias for C<$c-E<gt>tx-E<gt>res>.
-Usually refers to a L<Mojo::Message::Response> object.
+Alias for C<$c-E<gt>tx-E<gt>res>. Usually refers to a
+L<Mojo::Message::Response> object.
+
+  $c->res->headers->content_disposition('attachment; filename=foo.png;');
 
 =head2 C<respond_to>
 
@@ -910,7 +904,6 @@ Usually refers to a L<Mojo::Message::Response> object.
 Automatically select best possible representation for resource from C<Accept>
 request header, C<format> stash value or C<format> GET/POST parameter,
 defaults to rendering an empty C<204> response.
-Note that this method is EXPERIMENTAL and might change without warning!
 
   $c->respond_to(
     json => sub { $c->render_json({just => 'works'}) },
@@ -918,14 +911,29 @@ Note that this method is EXPERIMENTAL and might change without warning!
     any  => {data => '', status => 204}
   );
 
-=head2 C<send_message>
+=head2 C<send>
 
-  $c = $c->send_message('Hi there!');
-  $c = $c->send_message('Hi there!', sub {...});
+  $c = $c->send({binary => $bytes});
+  $c = $c->send({text   => $bytes});
+  $c = $c->send([$fin, $rsv1, $rsv2, $rsv3, $op, $payload]);
+  $c = $c->send('Hi there!');
+  $c = $c->send('Hi there!', sub {...});
 
-Send a message non-blocking via WebSocket, the optional drain callback will
-be invoked once all data has been written.
-Note that this method is EXPERIMENTAL and might change without warning!
+Send message or frame non-blocking via WebSocket, the optional drain callback
+will be invoked once all data has been written. Note that this method is
+EXPERIMENTAL and might change without warning!
+
+  # Send JSON object as text frame
+  $c->send({text => Mojo::JSON->new->encode({hello => 'world'})});
+
+  # Send "Ping" frame
+  $c->send([1, 0, 0, 0, 9, 'Hello World!']);
+
+For mostly idle WebSockets you might also want to increase the inactivity
+timeout, which usually defaults to C<15> seconds.
+
+  # Increase inactivity timeout for connection to 300 seconds
+  Mojo::IOLoop->stream($c->tx->connection)->timeout(300);
 
 =head2 C<session>
 
@@ -934,8 +942,8 @@ Note that this method is EXPERIMENTAL and might change without warning!
   $c          = $c->session({foo => 'bar'});
   $c          = $c->session(foo => 'bar');
 
-Persistent data storage, defaults to using signed cookies.
-Note that cookies are generally limited to 4096 bytes of data.
+Persistent data storage, stored C<JSON> serialized in a signed cookie. Note
+that cookies are generally limited to 4096 bytes of data.
 
   $c->session->{foo} = 'bar';
   my $foo = $c->session->{foo};
@@ -959,7 +967,11 @@ Cookies failing signature verification will be automatically discarded.
   $c        = $c->stash(foo => 'bar');
 
 Non persistent data storage and exchange, application wide default values can
-be set with L<Mojolicious/"defaults">.
+be set with L<Mojolicious/"defaults">. Many stash value have a special
+meaning and are reserved, the full list is currently C<action>, C<app>,
+C<cb>, C<controller>, C<data>, C<extends>, C<format>, C<handler>, C<json>,
+C<layout>, C<namespace>, C<partial>, C<path>, C<status>, C<template> and
+C<text>.
 
   $c->stash->{foo} = 'bar';
   my $foo = $c->stash->{foo};
@@ -969,8 +981,8 @@ be set with L<Mojolicious/"defaults">.
 
   my $ua = $c->ua;
 
-Alias for C<$c-E<gt>app-E<gt>ua>.
-Usually refers to a L<Mojo::UserAgent> object.
+Alias for C<$c-E<gt>app-E<gt>ua>. Usually refers to a L<Mojo::UserAgent>
+object.
 
   # Blocking
   my $tx = $c->ua->get('http://mojolicio.us');
@@ -1005,11 +1017,17 @@ Usually refers to a L<Mojo::UserAgent> object.
 
 Generate a portable L<Mojo::URL> object with base for a route, path or URL.
 
-  # "/perldoc" if application is deployed under "/"
-  say $c->url_for('/perldoc');
+  # "/perldoc?foo=bar" if application is deployed under "/"
+  $c->url_for('/perldoc')->query(foo => 'bar');
 
-  # "/myapp/perldoc" if application is deployed under "/myapp"
-  say $c->url_for('/perldoc');
+  # "/myapp/perldoc?foo=bar" if application is deployed under "/myapp"
+  $c->url_for('/perldoc')->query(foo => 'bar');
+
+You can also use L<Mojolicious::Plugin::DefaultHelpers/"url_with"> to inherit
+query parameters from the current request.
+
+  # "/list?q=mojo&page=2" if current request was for "/list?q=mojo&page=1"
+  $c->url_with->query([page => 2]);
 
 =head2 C<write>
 
@@ -1037,11 +1055,11 @@ invoked once all data has been written.
     });
   });
 
-For Comet (C<long polling>) you might also want to increase the connection
+For Comet (C<long polling>) you might also want to increase the inactivity
 timeout, which usually defaults to C<15> seconds.
 
-  # Increase timeout for current connection to 300 seconds
-  Mojo::IOLoop->connection_timeout($c->tx->connection => 300);
+  # Increase inactivity timeout for connection to 300 seconds
+  Mojo::IOLoop->stream($c->tx->connection)->timeout(300);
 
 =head2 C<write_chunk>
 
@@ -1050,7 +1068,7 @@ timeout, which usually defaults to C<15> seconds.
   $c->write_chunk(sub {...});
   $c->write_chunk('Hello!', sub {...});
 
-Write dynamic content non-blocking with the C<chunked> transfer encoding, the
+Write dynamic content non-blocking with C<chunked> transfer encoding, the
 optional drain callback will be invoked once all data has been written.
 
   $c->write_chunk('He', sub {
@@ -1074,8 +1092,8 @@ You can call C<finish> at any time to end the stream.
 =head1 HELPERS
 
 In addition to the attributes and methods above you can also call helpers on
-instances of L<Mojolicious::Controller>.
-This includes all helpers from L<Mojolicious::Plugin::DefaultHelpers> and
+instances of L<Mojolicious::Controller>. This includes all helpers from
+L<Mojolicious::Plugin::DefaultHelpers> and
 L<Mojolicious::Plugin::TagHelpers>.
 
   $c->layout('green');
